///////////////////////////////////////////
// Processingによる二分法の対話的プログラム
//                      2012/04/28 T.Shirai
//             shirai＠mech.suzuka-ct.ac.jp
///////////////////////////////////////////

void Options() {
  crowbar.fontSize(18);
  crowbar.bgColor(#000000);
  crowbar.systemColor(#ffffff, #ff0000, #8888ff);
  crowbar.startLogging();
}

void initCrowbar() {
  startCrowbar.generate(60, 30);
}

void Setup() {
  crowbar.programComment("二分法のプログラム");
  crowbar.programComment("区間aは区間bよりも小さな値（例： a = 1, b = 2）");

  crowbar.param("区間ａ").setFloat(1.0).label("xa");
  crowbar.param("区間ｂ").setFloat(2.0).label("xb");
}

// メインルーチン
void Main() {
  float err;
  float a, b, c, tmp;
  int   loopcount;

  // 許容誤差
  err = 0.000001;
  // 変数の代入
  a = crowbar.getFloat("xa");
  b = crowbar.getFloat("xb");
  // 範囲のチェックその１
  if (a > b) {
    tmp = a; a = b; b = tmp; // swap
  }
  // 範囲のチェックその２
  if (sgn(f(a)) == sgn(f(b))) {
    crowbar.textColor(#ff0000).writeln("Error: 値の範囲が正しくありません．").writeln("f(a) = " + nf(f(a), 1, 4) + " f(b) = " + nf(f(b), 1, 3));
    return;
  }
  // 繰り返し計算
  loopcount = 0;
  displayBSM(loopcount, a, b);
  while (abs(b - a) >= err) {
    c = (a + b) / 2.0;
    if (sgn(f(a)) == sgn(f(c))) a = c; else b = c;
    loopcount++;
    displayBSM(loopcount, a, b);
  }
  crowbar.textColor(#00ff00);
  // 最終結果
  crowbar.newline().textColor(#ffffff).writeln(str(loopcount) + "回で収束しました．").writeln("結果は" + nf((a + b) / 2.0, 1, 10) + "です．");
}

// 問題の方程式
float f(float x) {
  float y;
  y = pow(x,3) - 3.0 * pow(x, 2) + 9.0 * x - 8.0;
  return y;
}

// 区間aと区間bの値を表示する関数
void displayBSM(int i, float a, float b) {
  crowbar.textColor(#ffffff).write("[" + nf(i, 3) + "] ");
  crowbar.textColor(#00ffff).write("区間A : xa = " + nf(a, 1, 10) + " f(xa) = " + nf(f(a), 1, 10));
  crowbar.textColor(#ffffff).write(" / ");
  crowbar.textColor(#ff00ff).write("区間B : xb = " + nf(b, 1, 10) + " f(xb) = " + nf(f(b), 1, 10));
  crowbar.newline();
}
